class x86_Linux:
    @staticmethod
    def convert_to_asm(data:str, CompiledLangSet:dict, CompiledLang:dict) -> tuple[str, dict, int]:
        Output_data = ""
        Data_SECTION = {}

        data_lines = data.splitlines()
        temp_lines = ""

        section_text = "section .text\nglobal _start\n"
        section_data = "section .data\n"
        section_bss = "section .bss\n"

        func_list = []

        _start = "_start:\n"

        Include = False
        IncludeName = ""

        for i, v in enumerate(data_lines):
            data_words = v.split(" ")
            
            for index, word in enumerate(data_words):
                DVALUE = ""
                DSIZE = ""
                DNAME = ""

                if Include:
                    CompiledLang["SET"] = CompiledLang["SET"][IncludeName]
                    Include = False

                for index_, INS in enumerate(CompiledLang["SET"].keys()):

                    if CompiledLang["SET"][INS] == "includeFuncs" and "/INCLUDEINS" in INS:
                        split_INS = ''
                        seperator = ''
                        split_word = ''

                        INS = INS.replace("/INCLUDEINS", "")

                        if INS.split("/S")[1] == "":
                            split_INS = INS
                            seperator = None
                            split_word = word
                        else:
                            seperator = INS.split("/S")[1]
                            split_INS = INS.replace("/S", "").replace(seperator, "").split("/*")
                            split_word = data_lines[i].split(seperator)

                        if split_INS[0] in word:
                            for point_index, point_ in enumerate(split_INS):

                                if point_index == 0:
                                    continue

                                if point_ == "DVALUE":
                                    DVALUE = split_word[point_index]

                            for _Index, _Value in enumerate(CompiledLangSet["funcs"][DVALUE]):
                                func_list.append(_Value)

                            IncludeName = DVALUE + "/INCLUDE"
                            DVALUE = ""
                            Include = True
                            break
                        else:
                            raise ValueError("INVALID INCLUDE INSTRUCTION")

                    elif INS.split("/*")[0].replace("/S", "") in word and not "/INCLUDE" in INS:
                        word_s = ''

                        if INS.split("/S")[1] == "":
                            word_s = word
                        else:
                            word_s = v.split(INS.split("/S")[1])

                        INS_ = INS

                        INS = INS.replace("/S", "").replace(" ", "")

                        for index__, point in enumerate(INS.split("/*")):
                            if index__ == 0:
                                pass
                            else:
                                if point == "DVALUE":
                                    DVALUE = word_s[index__]
                                elif point == "DNAME":
                                    DNAME = word_s[index__]
                                elif point == "DSIZE":
                                    DSIZE = word_s[index__]
                                else:
                                    pass
                        
                        if CompiledLang["SET"][INS_] in CompiledLangSet["INSTRUCTION_SET"].keys():
                            _start += CompiledLangSet["INSTRUCTION_SET"][CompiledLang["SET"][INS_]].replace("*%*DNAME*%*", DNAME).replace("*%*DVALUE*%*", DVALUE).replace("*%*DSIZE*%*", DSIZE)
                            if CompiledLang["SET"][INS_] in CompiledLangSet["DATA_SET"].keys():
                                section_data += CompiledLangSet["DATA_SET"][CompiledLang["SET"][INS_]].replace("*%*DNAME*%*", DNAME).replace("*%*DVALUE*%*", DVALUE).replace("*%*DSIZE*%*", DSIZE)
                            
                            if CompiledLang["SET"][INS_] in CompiledLangSet["BSS_SET"].keys():
                                section_bss += CompiledLangSet["BSS_SET"][CompiledLang["SET"][INS_]].replace("*%*DNAME*%*", DNAME).replace("*%*DVALUE*%*", DVALUE).replace("*%*DSIZE*%*", DSIZE)

        _start += CompiledLangSet["S_TEXT_END"]

        func_out = ""

        if _start == "_start:\n":
            raise Exception("Compiled Code doesn't have any Instructions!")

        for i, v in enumerate(func_list):
            func_out += v + "\n"

        # Process Data Section for variable definitions
        for _index, line in enumerate(section_data.splitlines()):
            if _index == 0:
                continue

            if "equ" in line:
                var_name, value_expr = line.split(" equ ")
                var_name = var_name.strip()
                value = ""
                # Calculate the value (currently assumes linear addressing)
                if "$-" in value_expr.replace(" ", ""):
                    value = len(Data_SECTION[value_expr.replace(" ", "").replace("$-", "")])
                else:
                    value = value_expr
                Data_SECTION[var_name] = value
            elif "db" in line:
                Data_SECTION[line.split(' db ')[0]] = line.split(" db ")[1].replace("\"", "")
            elif var_name := line.split()[0]:
                # Handle direct value assignments
                Data_SECTION[var_name] = line

        # Process BSS Section for variable definitions
        for __index, line in enumerate(section_bss.splitlines()):
            if __index == 0:
                continue

            if "resb" in line:
                var_name, value_expr = line.split(" resb ")
                var_name = var_name.strip()
                value = value_expr
                Data_SECTION[var_name] = value
        
        cancelled_sections = 0

        if section_bss != "section .bss\n":
            if section_data != "section .data\n":
                Output_data = f"{section_data}\n{section_bss}\n{section_text}\n{_start}\n{func_out}"
            else:
                Output_data = f"{section_bss}\n{section_text}\n{_start}\n{func_out}"
                cancelled_sections += 1
        else:
            cancelled_sections += 1
            if section_data != "section .data\n":
                Output_data = f"{section_data}\n{section_text}\n{_start}\n{func_out}"
            else:
                Output_data = f"{section_text}\n{_start}\n{func_out}"
                cancelled_sections += 1

        return Output_data, Data_SECTION, cancelled_sections


import os

class Paths:
    # NOEXT Files are to be made without the use of the foldername.
    #Folders
    phardwareitk = os.path.dirname(os.path.abspath(__file__))
    ErrorSystem = os.path.join(phardwareitk, 'ErrorSystem')
    Extensions = os.path.join(phardwareitk, 'Extensions')
    FileSystem = os.path.join(phardwareitk, 'FileSystem')
    GUI = os.path.join(phardwareitk, 'GUI')
    HGame = os.path.join(phardwareitk, 'HGame')
    ModuleController = os.path.join(phardwareitk, 'ModuleController')
    PLTEC = os.path.join(phardwareitk, 'PLTEC')
    System = os.path.join(phardwareitk, 'System')
    CLI = os.path.join(phardwareitk, 'CLI')

    C = os.path.join(GUI, 'C')

    #Inits
    __init__PY_phardwareitk = os.path.join(phardwareitk, '__init__.py')
    __init__PY_ErrorSystem = os.path.join(ErrorSystem, '__init__.py')
    __init__PY_Extensions = os.path.join(Extensions, '__init__.py')
    __init__PY_FileSystem = os.path.join(FileSystem, '__init__.py')
    __init__PY_GUI = os.path.join(GUI, '__init__.py')
    __init__PY_HGame = os.path.join(HGame, '__init__.py')
    __init__PY_ModuleController = os.path.join(ModuleController, '__init__.py')
    __init__PY_PLTEC = os.path.join(PLTEC, '__init__.py')
    __init__PY_System = os.path.join(System, '__init__.py')
    __init__PY_CLI = os.path.join(CLI, '__init__.py')

    #Files in phardwareitk
    defrunSettingsNOEXT = os.path.join(phardwareitk, 'defrunSettings')
    DependenciesPY_phardwareitk = os.path.join(phardwareitk, 'Dependencies.py')

    #Files in CLI
    cliToolKitPY_CLI = os.path.join(CLI, 'cliToolKit.py')

    #Files in ErrorSystem
    ErrorSystemPY_ErrorSystem = os.path.join(ErrorSystem, 'ErrorSystem.py')

    #Files in Extention
    HyperOutPY_Extensions = os.path.join(Extensions, 'HyperOut.py')
    HyperInPY_Extensions = os.path.join(Extensions, 'HyperIn.py')

    #Files in FileSystem
    FileSystemPY_FileSystem = os.path.join(FileSystem, 'FileSystem.py')

    #Files in GUI
    guiPY_GUI = os.path.join(GUI, 'gui.py')
    gui_sdlPY_GUI = os.path.join(GUI, 'gui_sdl.py')
    renderGUIPYX_GUI = os.path.join(GUI, 'renderGUI.pyx')
    renderGUIPYD_GUI = os.path.join(GUI, 'renderGUI.pyd')
    renderGUISO_GUI = os.path.join(GUI, 'renderGUI.so')

    #Files in HGame

    #Files in ModuleController
    mainPY_ModuleController = os.path.join(ModuleController, 'main.py')

    #Files in PLTEC
    anayzerPY_PLTEC = os.path.join(PLTEC, 'anayzer.py')
    ASMPY_PLTEC = os.path.join(PLTEC, 'ASM.py')
    CheckerPY_PLTEC = os.path.join(PLTEC, 'Checker.py')
    DEF_langJSON_PLTEC = os.path.join(PLTEC, 'DEF_lang.json')
    DEF_LanguageSETJSON_PLTEC = os.path.join(PLTEC, 'DEF_LanguageSET.json')
    LinkerPY_PLTEC = os.path.join(PLTEC, 'Linker.py')
    LoggerPY_PLTEC = os.path.join(PLTEC, 'Logger.py')
    mainPY_PLTEC = os.path.join(PLTEC, 'main.py')
    OBJECTPY_PLTEC = os.path.join(PLTEC, 'OBJECT.py')
    ReaderPY_PLTEC = os.path.join(PLTEC, 'Reader.py')

    #File in System
    SysUsagePY_System = os.path.join(System, 'SysUsage.py')


import os
import sys
import binascii
import json

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.backends import default_backend

from . import *

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk import LIB

import phardwareitk.ErrorSystem.ErrorSystem as ES

from typing import *

class BasicFileSystem:
    @staticmethod
    def CreateFile(filePath:str, data:Union[str, dict, int]) -> tuple[bool, str]:
        """Creates a new file with the provided data.
        
        Args:
            filePath (str): The path to the file to be created.
            date (Union[str, dict, int]) : The data to be inserted into the file.
            
        Returns:
            tuple[bool, str] ->
                bool: True if successful, False otherwise.
                str: The Error if success was False.
            
        Calls (Without raising):
            PheonixException: If an Unknown Error occurs during creation.
            PheonixIOError: If an Input/Output Error occurs during creation."""
        success:bool = False
        error:str = ""
        
        try:
            with open(filePath, 'w') as f:
                f.write(str(data))
                f.close()

            error = ""
            success = True
        except IOError as e:
            error = ES.PheonixIOError(str(e))
            success = False
            return success, error
        except Exception as e:
            error = ES.PheonixException(str(e))
            success = False
            return success, error

        return success, error

    @staticmethod
    def AppendFile(filePath:str, data:Union[str, int, dict], seek:Union[None, int]=None) -> tuple[bool, str]:
        """Appends the data to the given filePath

        Args:
            filePath (str): The path to the file to append.
            data (Union[str, int, dict]): The data to append in the file.
            seek (Optional) (Union[None, int]): The seek position. 'None' means to append at the end of the file. Defaults to None.

        Returns:
            tuple[bool, str] ->
                bool: True if successful, False otherwise.
                str: The Error if success was False

        Calls (Without raising):
            PheonixException: If an Unknown Error occurs during creation.
            PheonixIOError: If an Input/Output Error occurs during creation.
            PheonixFileNotFoundError: If file was not found."""
        success:bool = False
        error:str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File doesn't exist [{filePath}]")
            success = False
            return success, error
        
        else:
            try:
                if seek != None:
                    with open(filePath, "a") as f:
                        f.seek(seek)
                        f.write(str(data))
                        f.close()
                else:
                    with open(filePath, "a") as f:
                        f.write(str(data))
                        f.close()

                error = ""
                success = True
            except IOError as e:
                error = ES.PheonixIOError(str(e))
                success = False
                return success, error
            except Exception as e:
                error = ES.PheonixException(str(e))
                success = False
                return success, error
            
            return success, error
        
    @staticmethod
    def ReadFile(filePath: str) -> tuple[bool, Union[str, None]]:
        """Reads the content of the file.

        Args:
            filePath (str): The path to the file to read.

        Returns:
            tuple[bool, Union[str, None]] -> 
                bool: True if successful, False otherwise.
                str: The content of the file if successful, or error message if not.

        Calls (Without raising):
            PheonixException: If an Unknown Error occurs during reading.
            PheonixFileNotFoundError: If file was not found."""
        success: bool = False
        content: Union[str, None] = None
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File doesn't exist [{filePath}]")
            success = False
            return success, error
        
        try:
            with open(filePath, 'r') as f:
                content = f.read()
            error = content
            success = True
        except IOError as e:
            error = ES.PheonixIOError(str(e))
            success = False
        except Exception as e:
            error = ES.PheonixException(str(e))
            success = False

        return success, error

    @staticmethod
    def DeleteFile(filePath: str) -> tuple[bool, str]:
        """Deletes the specified file.

        Args:
            filePath (str): The path to the file to delete.

        Returns:
            tuple[bool, str] ->
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file was not found."""
        success: bool = False
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File doesn't exist [{filePath}]")
            success = False
            return success, error

        try:
            os.remove(filePath)
            success = True
            error = ""
        except IOError as e:
            error = ES.PheonixIOError(str(e))
            success = False
        except Exception as e:
            error = ES.PheonixException(str(e))
            success = False

        return success, error

    @staticmethod
    def RenameFile(oldPath: str, newPath: str) -> tuple[bool, str]:
        """Renames a file from `oldPath` to `newPath`.

        Args:
            oldPath (str): The current path of the file to rename.
            newPath (str): The new name/path of the file.

        Returns:
            tuple[bool, str] ->
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file was not found."""
        success: bool = False
        error: str = ""

        if not os.path.exists(oldPath):
            error = ES.PheonixFileNotFoundError(f"File doesn't exist [{oldPath}]")
            success = False
            return success, error

        try:
            os.rename(oldPath, newPath)
            success = True
            error = ""
        except IOError as e:
            error = ES.PheonixIOError(str(e))
            success = False
        except Exception as e:
            error = ES.PheonixException(str(e))
            success = False

        return success, error

    @staticmethod
    def FileExists(filePath: str) -> tuple[bool, str]:
        """Checks if the file exists at the specified path.

        Args:
            filePath (str): The path to check.

        Returns:
            tuple[bool, str] ->
                bool: True if the file exists, False otherwise.
                str: An appropriate message."""
        exists: bool = False
        error: str = ""

        if os.path.exists(filePath):
            exists = True
            error = "File exists."
        else:
            exists = False
            error = ES.PheonixFileNotFoundError(f"File not found at [{filePath}]")

        return exists, error

    @staticmethod
    def GetFileSize(filePath: str) -> tuple[bool, Union[int, str]]:
        """Returns the size of the specified file.

        Args:
            filePath (str): The path to the file.

        Returns:
            tuple[bool, Union[int, str]] ->
                bool: True if successful, False otherwise.
                int: The size of the file in bytes if successful.
                str: Error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file was not found."""
        success: bool = False
        error: Union[int, str] = 0

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File doesn't exist [{filePath}]")
            success = False
            return success, error

        try:
            size = os.path.getsize(filePath)
            error = size
            success = True
        except IOError as e:
            error = ES.PheonixIOError(str(e))
            success = False
        except Exception as e:
            error = ES.PheonixException(str(e))
            success = False

        return success, error

    @staticmethod
    def CopyFile(srcPath: str, destPath: str) -> tuple[bool, str]:
        """Copies a file from the source path to the destination path.

        Args:
            srcPath (str): The source file path.
            destPath (str): The destination file path.

        Returns:
            tuple[bool, str] ->
                bool: True if successful, False otherwise.
                str: Error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If the source file does not exist."""
        success: bool = False
        error: str = ""

        if not os.path.exists(srcPath):
            error = ES.PheonixFileNotFoundError(f"Source file doesn't exist [{srcPath}]")
            success = False
            return success, error

        try:
            import shutil
            shutil.copy(srcPath, destPath)
            success = True
            error = ""
        except IOError as e:
            error = ES.PheonixIOError(str(e))
            success = False
        except Exception as e:
            error = ES.PheonixException(str(e))
            success = False

        return success, error

    @staticmethod
    def MoveFile(srcPath: str, destPath: str) -> tuple[bool, str]:
        """Moves a file from the source path to the destination path.

        Args:
            srcPath (str): The source file path.
            destPath (str): The destination file path.

        Returns:
            tuple[bool, str] ->
                bool: True if successful, False otherwise.
                str: Error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If the source file does not exist."""
        success: bool = False
        error: str = ""

        if not os.path.exists(srcPath):
            error = ES.PheonixFileNotFoundError(f"Source file doesn't exist [{srcPath}]")
            success = False
            return success, error

        try:
            os.rename(srcPath, destPath)
            success = True
            error = ""
        except IOError as e:
            error = ES.PheonixIOError(str(e))
            success = False
        except Exception as e:
            error = ES.PheonixException(str(e))
            success = False

        return success, error

    @staticmethod
    def CreateDirectory(dirPath: str) -> tuple[bool, str]:
        """Creates a directory at the specified path.

        Args:
            dirPath (str): The path where the directory will be created.

        Returns:
            tuple[bool, str] ->
                bool: True if successful, False otherwise.
                str: Error message if unsuccessful.
        """
        success: bool = False
        error: str = ""

        try:
            os.makedirs(dirPath, exist_ok=True)
            success = True
            error = ""
        except IOError as e:
            error = ES.PheonixIOError(str(e))
            success = False
        except Exception as e:
            error = ES.PheonixException(str(e))
            success = False

        return success, error

class JsonFileSystem:
    @staticmethod
    def CreateJsonFile(filePath: str, data: Union[dict, list]) -> tuple[bool, str]:
        """Creates a new JSON file with the provided data.
        
        Args:
            filePath (str): The path to the JSON file to be created.
            data (Union[dict, list]): The data to be inserted into the JSON file.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.
            
        Calls (Without raising):
            PheonixException: If an unknown error occurs during file creation.
            PheonixIOError: If there is an input/output error during file creation."""
        
        import json
        success: bool = False
        error: str = ""

        try:
            with open(filePath, 'w') as f:
                json.dump(data, f, indent=4)
            success = True
        except IOError as e:
            error = ES.PheonixIOError(str(e))
        except Exception as e:
            error = ES.PheonixException(str(e))
        
        return success, error

    @staticmethod
    def AppendJsonData(filePath: str, data: Union[dict, list]) -> tuple[bool, str]:
        """Appends data to an existing JSON file.

        Args:
            filePath (str): The path to the JSON file to append data.
            data (Union[dict, list]): The data to be appended.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: Error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If the file is not found.
            PheonixIOError: If an input/output error occurs."""
        success: bool = False
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File not found: {filePath}")
        else:
            try:
                with open(filePath, 'r+') as f:
                    current_data = json.load(f)
                    if isinstance(current_data, list):
                        current_data.append(data)
                    elif isinstance(current_data, dict):
                        current_data.update(data)
                    else:
                        error = ES.PheonixException("Unsupported data format in file.")
                        return success, error
                    f.seek(0)
                    json.dump(current_data, f, indent=4)
                success = True
            except IOError as e:
                error = ES.PheonixIOError(str(e))
            except Exception as e:
                error = ES.PheonixException(str(e))

        return success, error

    @staticmethod
    def ReadJsonFile(filePath: str) -> tuple[bool, Union[dict, list, str]]:
        """Reads data from a JSON file.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, Union[dict, list, str]] -> 
                bool: True if successful, False otherwise.
                dict/list: The contents of the file if successful, error message if not.

        Calls (Without raising):
            PheonixFileNotFoundError: If file not found.
            PheonixIOError: If an input/output error occurs."""
        success: bool = False
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File not found: {filePath}")
        else:
            try:
                with open(filePath, 'r') as f:
                    data = json.load(f)
                    success = True
                    error = data
            except IOError as e:
                error = ES.PheonixIOError(str(e))
            except json.JSONDecodeError as e:
                error = ES.PheonixException(f"JSON decoding error: {str(e)}")
            except Exception as e:
                error = ES.PheonixException(str(e))

        return success, error

    @staticmethod
    def DeleteJsonFile(filePath: str) -> tuple[bool, str]:
        """Deletes a JSON file.

        Args:
            filePath (str): The path to the JSON file to be deleted.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file does not exist."""
        success: bool = False
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File not found: {filePath}")
        else:
            try:
                os.remove(filePath)
                success = True
            except IOError as e:
                error = ES.PheonixIOError(str(e))
            except Exception as e:
                error = ES.PheonixException(str(e))

        return success, error

    @staticmethod
    def FileExists(filePath: str) -> bool:
        """Checks if the JSON file exists.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            bool: True if the file exists, False otherwise."""
        return os.path.exists(filePath)

    @staticmethod
    def GetJsonKeys(filePath: str) -> tuple[bool, Union[list, str]]:
        """Returns all keys in a JSON file (if it is a JSON object).

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, Union[list, str]] -> 
                bool: True if successful, False otherwise.
                list: List of keys if file is a dictionary, error message if not.

        Calls (Without raising):
            PheonixFileNotFoundError: If file does not exist."""
        success: bool = False
        error: str = []

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File not found: {filePath}")
        else:
            try:
                with open(filePath, 'r') as f:
                    data = json.load(f)
                    if isinstance(data, dict):
                        success = True
                        error = list(data.keys())
                    else:
                        error = ES.PheonixException("File does not contain a JSON object.")
            except IOError as e:
                error = ES.PheonixIOError(str(e))
            except json.JSONDecodeError as e:
                error = ES.PheonixException(f"JSON decoding error: {str(e)}")
            except Exception as e:
                error = ES.PheonixException(str(e))

        return success, error

    @staticmethod
    def GetJsonFileSize(filePath: str) -> tuple[bool, Union[int, str]]:
        """Returns the size of the JSON file.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, Union[int, str]] -> 
                bool: True if successful, False otherwise.
                int: File size in bytes if successful, error message if not.

        Calls (Without raising):
            PheonixFileNotFoundError: If file does not exist."""
        success: bool = False
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File not found: {filePath}")
        else:
            try:
                success = True
                error = os.path.getsize(filePath)
            except Exception as e:
                error = ES.PheonixException(str(e))

        return success, error

    @staticmethod
    def UpdateJsonKey(filePath: str, key: str, new_value: Union[str, int, dict, list]) -> tuple[bool, str]:
        """Updates the value of a specific key in a JSON file.

        Args:
            filePath (str): The path to the JSON file.
            key (str): The key whose value needs to be updated.
            new_value (Union[str, int, dict, list]): The new value to update.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file does not exist."""
        success: bool = False
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File not found: {filePath}")
        else:
            try:
                with open(filePath, 'r+') as f:
                    data = json.load(f)
                    if isinstance(data, dict) and key in data:
                        data[key] = new_value
                        f.seek(0)
                        json.dump(data, f, indent=4)
                        success = True
                    else:
                        error = ES.PheonixException(f"Key '{key}' not found in JSON.")
            except IOError as e:
                error = ES.PheonixIOError(str(e))
            except json.JSONDecodeError as e:
                error = ES.PheonixException(f"JSON decoding error: {str(e)}")
            except Exception as e:
                error = ES.PheonixException(str(e))

        return success, error

    @staticmethod
    def ValidateJson(filePath: str) -> tuple[bool, str]:
        """Validates if a file is a valid JSON.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, str] -> 
                bool: True if valid, False otherwise.
                str: The error message if invalid.

        Calls (Without raising):
            PheonixFileNotFoundError: If file not found."""
        success: bool = False
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File not found: {filePath}")
        else:
            try:
                with open(filePath, 'r') as f:
                    json.load(f)
                success = True
            except json.JSONDecodeError as e:
                error = ES.PheonixException(f"Invalid JSON: {str(e)}")
            except Exception as e:
                error = ES.PheonixException(str(e))

        return success, error

    @staticmethod
    def GetJsonFileExtension(filePath: str) -> tuple[bool, str]:
        """Gets the file extension of a JSON file.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The file extension ('.json') if successful, error message if not."""
        success: bool = False
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File not found: {filePath}")
        else:
            _, ext = os.path.splitext(filePath)
            if ext == ".json":
                success = True
                error = ext
            else:
                error = ES.PheonixException(f"File is not a JSON file: {filePath}")

        return success, error

    @staticmethod
    def MergeJsonFiles(srcPath: str, destPath: str) -> tuple[bool, str]:
        """Merges two JSON files.

        Args:
            srcPath (str): The source JSON file path.
            destPath (str): The destination JSON file path.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If any file is not found."""
        success: bool = False
        error: str = ""

        if not os.path.exists(srcPath):
            error = ES.PheonixFileNotFoundError(f"Source file not found: {srcPath}")
        elif not os.path.exists(destPath):
            error = ES.PheonixFileNotFoundError(f"Destination file not found: {destPath}")
        else:
            try:
                with open(srcPath, 'r') as f_src, open(destPath, 'r+') as f_dest:
                    src_data = json.load(f_src)
                    dest_data = json.load(f_dest)

                    if isinstance(src_data, dict) and isinstance(dest_data, dict):
                        dest_data.update(src_data)
                    else:
                        error = ES.PheonixException("Both files should contain JSON objects.")
                        return success, error

                    f_dest.seek(0)
                    json.dump(dest_data, f_dest, indent=4)
                    success = True
            except Exception as e:
                error = ES.PheonixException(str(e))

        return success, error

    @staticmethod
    def CreateJsonBackup(filePath: str) -> tuple[bool, str]:
        """Creates a backup of a JSON file.

        Args:
            filePath (str): The path to the JSON file to be backed up.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful."""
        success: bool = False
        error: str = ""

        if not os.path.exists(filePath):
            error = ES.PheonixFileNotFoundError(f"File not found: {filePath}")
        else:
            try:
                backup_path = filePath + ".bak"
                with open(filePath, 'r') as f_src, open(backup_path, 'w') as f_bak:
                    json.dump(json.load(f_src), f_bak, indent=4)
                success = True
            except Exception as e:
                error = ES.PheonixException(str(e))

        return success, error

class LowFileSystem:
    """NOTE: For using majority of the functions in this class, please install GCC, LD, and NASM (Mingw64 or Msys2 UCRT64).
    If you cannot install it. Please use PLTEC via Command Line Interface to convert any language syntax defined in a json file to Assembly or even Object (Undergoing development). But before making such a file please understand the file's syntax ->
        1. Include the convertion Os (ALL CAPS) and mode at the start of the json. Like -> 'ELF': '', 'x86': ''
        2. define a json inside the file with the key as 'SET'. Like -> '{'ELF': '', 'x86': '', 'SET': {}}'. The set is the point where all the syntax will go.
        3. NOTE: The syntax will work with the pre-made funcs in the DEF_LanguageSET.json. The syntax will have to be understood by the you if you want to add more funcs in it.
        4. Each Key in the 'SET' will have to have these required notes ->
            i. No spaces are allowed in the key
            ii. Include /S at the end of the key. And after the /S please include a sperator that will be used to seperate the key. Commomly used -> ' '.
            iii. Here are some important definitions ->
                a. /*DNAME: Specifies that at the specific location the user must type in the name of [something].
                b. /*DVALUE: Specifies that at the specific location the user must type in the value of [something].
                c. /*DSIZE: Specifies that at the specific location the user must type in the size of [something]. NOTE: DSIZE is a must in input function
        5. Inside the 'SET' please include a key:value pair with the value as 'includeFuncs'. This ensures all the neccarry functions will be imported if the user wrote the specific key.
        6. After that include whatever key you want as to be working with the syntax but please remember the 4.iii section. The definitions act as a parameter to the funcs and please openUP DEF_LanguageSET to see the parameters and funcs
        7. In the Value input the func you want to call if the specified key is displayed.

        EXAMPLE Lang.json -> '
        {
            "ELF": "",
            "x86": "",
            "SET": {
                "#include:/*DVALUE/INCLUDEINS/S ": "includeFuncs",
                "BasicIO/INCLUDE": {
                    "print:/*DVALUE/*DNAME/S ": "printINS",
                    "input:/*DNAME/*DSIZE/S ": "inputINS",
                    "clean:/S": "cleanupINS",
                    "return:/*DVALUE/S ": "returnINS"
                }
            }
        }'

        Example Test.txt -> '
        #include: BasicIO
        
        print: helloWorld HelloVariable
        clean:
        return: 0'

        Converted Assembly -> '
        section .data
        HelloVariable_pri db \"helloWorld\"
        HelloVariable_pri_LEN equ $ - HelloVariable_pri

        section .text
        global _start

        _start:
        mov ecx, HelloVariable_pri
        mov edx, HelloVariable_pri_LEN
        call print
        call cleanup32
        mov ebx, 0
        call return

        print:
        mov eax, 4
        mov ebx, 1
        int 80h

        input:
        mov eax, 3
        mov ebx, 1
        int 80h

        return:
        mov eax, 1
        int 80h

        cleanup32:
        xor eax, eax
        xor ebx, ebx
        xor ecx, ecx
        xor edx, edx
        xor esi, esi
        xor ebp, ebp
        ret
        '

        PLTEC has been updated and some changes in the DEF_lang.json might not bee understood. so please try using a AI to understand via the example provided.
    """
    # 1. ASM File Handling
    @staticmethod
    def CreateAsmFile(filePath: str, asmContent: str) -> tuple[bool, str]:
        """Creates a new ASM file with the provided ASM content."""
        try:
            with open(filePath, 'w') as f:
                f.write(asmContent)
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def ReadAsmFile(filePath: str) -> tuple[bool, str]:
        """Reads content from an ASM file."""
        try:
            with open(filePath, 'r') as f:
                content = f.read()
            return True, content
        except Exception as e:
            return False, str(e)

    @staticmethod
    def CompileAsmToObject(filePath: str) -> tuple[bool, str]:
        """Compiles an ASM file to an object (.o) file."""
        try:
            os.system(f"nasm -f elf64 -o {filePath.replace('.asm', '.o')} {filePath}")
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def DisassembleAsmFile(filePath: str) -> tuple[bool, str]:
        """Disassembles an ASM file to inspect the assembly instructions."""
        try:
            disasm = os.popen(f"objdump -d {filePath}").read()
            return True, disasm
        except Exception as e:
            return False, str(e)

    @staticmethod
    def CreateAsmFromString(asmCode: str, outputFile: str) -> tuple[bool, str]:
        """Creates an ASM file from a string of ASM code."""
        try:
            with open(outputFile, 'w') as f:
                f.write(asmCode)
            return True, ""
        except Exception as e:
            return False, str(e)

    # 2. Object File (.O) Handling
    @staticmethod
    def CreateObjectFile(filePath: str) -> tuple[bool, str]:
        """Creates an object (.o) file from a C/C++ file."""
        try:
            os.system(f"gcc -c {filePath} -o {filePath.replace('.c', '.o')}")
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def LinkObjectFiles(objFiles: list[str], outputFile: str) -> tuple[bool, str]:
        """Links multiple object (.o) files into an executable."""
        try:
            os.system(f"ld -o {outputFile} {' '.join(objFiles)}")
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def DisassembleObjectFile(objFilePath: str) -> tuple[bool, str]:
        """Disassembles an object file to inspect the assembly instructions."""
        try:
            disasm = os.popen(f"objdump -d {objFilePath}").read()
            return True, disasm
        except Exception as e:
            return False, str(e)

    # 3. C and C++ File Handling
    @staticmethod
    def CreateCFile(filePath: str, cContent: str) -> tuple[bool, str]:
        """Creates a new C file with the provided C code."""
        try:
            with open(filePath, 'w') as f:
                f.write(cContent)
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def CompileCToObject(filePath: str) -> tuple[bool, str]:
        """Compiles a C file to an object (.o) file."""
        try:
            os.system(f"gcc -c {filePath} -o {filePath.replace('.c', '.o')}")
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def CreateCppFile(filePath: str, cppContent: str) -> tuple[bool, str]:
        """Creates a new C++ file with the provided C++ code."""
        try:
            with open(filePath, 'w') as f:
                f.write(cppContent)
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def CompileCppToObject(filePath: str) -> tuple[bool, str]:
        """Compiles a C++ file to an object (.o) file."""
        try:
            os.system(f"g++ -c {filePath} -o {filePath.replace('.cpp', '.o')}")
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def CompileCPlusPlusFile(filePath: str) -> tuple[bool, str]:
        """Compiles a C++ file into an executable."""
        try:
            os.system(f"g++ {filePath} -o {filePath.replace('.cpp', '')}")
            return True, ""
        except Exception as e:
            return False, str(e)

    # 4. Binary File Handling
    @staticmethod
    def ReadBinaryFile(filePath: str) -> tuple[bool, bytes]:
        """Reads data from a binary file."""
        try:
            with open(filePath, 'rb') as f:
                data = f.read()
            return True, data
        except Exception as e:
            return False, str(e)

    @staticmethod
    def WriteBinaryFile(filePath: str, data: bytes) -> tuple[bool, str]:
        """Writes data to a binary file."""
        try:
            with open(filePath, 'wb') as f:
                f.write(data)
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def AppendBinaryFile(filePath: str, data: bytes) -> tuple[bool, str]:
        """Appends data to a binary file."""
        try:
            with open(filePath, 'ab') as f:
                f.write(data)
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def GetBinaryFileSize(filePath: str) -> tuple[bool, int]:
        """Returns the size of a binary file."""
        try:
            size = os.path.getsize(filePath)
            return True, size
        except Exception as e:
            return False, str(e)

    # 5. Encryption and Decryption (using cryptography)
    @staticmethod
    def EncryptFile(filePath: str, key: bytes) -> tuple[bool, str]:
        """Encrypts the content of a file using AES encryption."""
        try:
            with open(filePath, 'rb') as f:
                data = f.read()

            # AES encryption setup
            cipher = Cipher(algorithms.AES(key), modes.CBC(os.urandom(16)), backend=default_backend())
            padder = padding.PKCS7(algorithms.AES.block_size).padder()
            padded_data = padder.update(data) + padder.finalize()
            encryptor = cipher.encryptor()
            encrypted_data = encryptor.update(padded_data) + encryptor.finalize()

            # Save encrypted file
            with open(filePath + ".enc", 'wb') as f_enc:
                f_enc.write(cipher.iv)  # Prepend IV for later decryption
                f_enc.write(encrypted_data)

            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def DecryptFile(filePath: str, key: bytes) -> tuple[bool, str]:
        """Decrypts an encrypted file."""
        try:
            with open(filePath, 'rb') as f:
                iv = f.read(16)  # Extract IV
                encrypted_data = f.read()

            # AES decryption setup
            cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend())
            decryptor = cipher.decryptor()
            decrypted_data = decryptor.update(encrypted_data) + decryptor.finalize()

            # Unpadding decrypted data
            unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
            unpadded_data = unpadder.update(decrypted_data) + unpadder.finalize()

            # Save decrypted file
            with open(filePath.replace('.enc', '.dec'), 'wb') as f_dec:
                f_dec.write(unpadded_data)

            return True, ""
        except Exception as e:
            return False, str(e)
        
    @staticmethod
    def Base64EncodeFile(filePath: str) -> tuple[bool, str]:
        """Encodes a file in Base64 format."""
        try:
            with open(filePath, 'rb') as f:
                encoded = binascii.b2a_base64(f.read())
            with open(filePath + '.b64', 'wb') as f_enc:
                f_enc.write(encoded)
            return True, ""
        except Exception as e:
            return False, str(e)

    @staticmethod
    def Base64DecodeFile(filePath: str) -> tuple[bool, str]:
        """Decodes a Base64 encoded file."""
        try:
            with open(filePath, 'rb') as f:
                decoded = binascii.a2b_base64(f.read())
            with open(filePath.replace('.b64', ''), 'wb') as f_dec:
                f_dec.write(decoded)
            return True, ""
        except Exception as e:
            return False, str(e)

class FileSystem(BasicFileSystem, JsonFileSystem, LowFileSystem):
    @staticmethod
    def CreateFile(filePath:str, data:str) -> tuple[bool, str]:
        """Creates a new file with the provided data.
        
        Args:
            filePath (str): The path to the file to be created.
            date (Union[str, dict, int]) : The data to be inserted into the file.
            
        Returns:
            tuple[bool, str] ->
                bool: True if successful, False otherwise.
                str: The Error if success was False.
            
        Calls (Without raising):
            PheonixException: If an Unknown Error occurs during creation.
            PheonixIOError: If an Input/Output Error occurs during creation."""
        return super(BasicFileSystem, BasicFileSystem).CreateFile(filePath, data)
    
    @staticmethod
    def AppendFile(filePath: str, data: Union[str, int, dict], seek: Optional[Union[None, int]] = None) -> tuple[bool, str]:
        """Appends the data to the given filePath.

        Args:
            filePath (str): The path to the file to append.
            data (Union[str, int, dict]): The data to append in the file.
            seek (Optional) (Union[None, int]): The seek position. 'None' means to append at the end of the file. Defaults to None.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The Error if success was False

        Calls (Without raising):
            PheonixException: If an Unknown Error occurs during creation.
            PheonixIOError: If an Input/Output Error occurs during creation.
            PheonixFileNotFoundError: If file was not found."""
        return super(BasicFileSystem, FileSystem).AppendFile(filePath, data, seek)

    @staticmethod
    def ReadFile(filePath: str) -> tuple[bool, Union[str, None]]:
        """Reads the content of the file.

        Args:
            filePath (str): The path to the file to read.

        Returns:
            tuple[bool, Union[str, None]] -> 
                bool: True if successful, False otherwise.
                str: The content of the file if successful, or error message if not.

        Calls (Without raising):
            PheonixException: If an Unknown Error occurs during reading.
            PheonixFileNotFoundError: If file was not found."""
        return super(BasicFileSystem, FileSystem).ReadFile(filePath)

    @staticmethod
    def DeleteFile(filePath: str) -> tuple[bool, str]:
        """Deletes the specified file.

        Args:
            filePath (str): The path to the file to delete.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file was not found."""
        return super(BasicFileSystem, FileSystem).DeleteFile(filePath)

    @staticmethod
    def RenameFile(oldPath: str, newPath: str) -> tuple[bool, str]:
        """Renames a file from `oldPath` to `newPath`.

        Args:
            oldPath (str): The current path of the file to rename.
            newPath (str): The new name/path of the file.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file was not found."""
        return super(BasicFileSystem, FileSystem).RenameFile(oldPath, newPath)

    @staticmethod
    def FileExists(filePath: str) -> tuple[bool, str]:
        """Checks if the file exists at the specified path.

        Args:
            filePath (str): The path to check.

        Returns:
            tuple[bool, str] -> 
                bool: True if the file exists, False otherwise.
                str: An appropriate message."""
        return super(BasicFileSystem, FileSystem).FileExists(filePath)

    @staticmethod
    def GetFileSize(filePath: str) -> tuple[bool, Union[int, str]]:
        """Returns the size of the specified file.

        Args:
            filePath (str): The path to the file.

        Returns:
            tuple[bool, Union[int, str]] -> 
                bool: True if successful, False otherwise.
                int: The size of the file in bytes if successful.
                str: Error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file was not found."""
        return super(BasicFileSystem, FileSystem).GetFileSize(filePath)

    @staticmethod
    def CopyFile(srcPath: str, destPath: str) -> tuple[bool, str]:
        """Copies a file from the source path to the destination path.

        Args:
            srcPath (str): The source file path.
            destPath (str): The destination file path.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: Error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If the source file does not exist."""
        return super(BasicFileSystem, FileSystem).CopyFile(srcPath, destPath)

    @staticmethod
    def MoveFile(srcPath: str, destPath: str) -> tuple[bool, str]:
        """Moves a file from the source path to the destination path.

        Args:
            srcPath (str): The source file path.
            destPath (str): The destination file path.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: Error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If the source file does not exist."""
        return super(BasicFileSystem, FileSystem).MoveFile(srcPath, destPath)

    @staticmethod
    def CreateDirectory(dirPath: str) -> tuple[bool, str]:
        """Creates a directory at the specified path.

        Args:
            dirPath (str): The path where the directory will be created.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: Error message if unsuccessful.
        """
        return super(BasicFileSystem, FileSystem).CreateDirectory(dirPath)
    
    @staticmethod
    def CreateJsonFile(filePath: str, data: Union[dict, list]) -> tuple[bool, str]:
        """Creates a new JSON file with the provided data.
        
        Args:
            filePath (str): The path to the JSON file to be created.
            data (Union[dict, list]): The data to be inserted into the JSON file.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixException: If an unknown error occurs during file creation.
            PheonixIOError: If there is an input/output error during file creation."""
        return super(JsonFileSystem, JsonFileSystem).CreateJsonFile(filePath, data)

    @staticmethod
    def AppendJsonData(filePath: str, data: Union[dict, list]) -> tuple[bool, str]:
        """Appends data to an existing JSON file.

        Args:
            filePath (str): The path to the JSON file to append data.
            data (Union[dict, list]): The data to be appended.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: Error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If the file is not found.
            PheonixIOError: If an input/output error occurs."""
        return super(JsonFileSystem, JsonFileSystem).AppendJsonData(filePath, data)

    @staticmethod
    def ReadJsonFile(filePath: str) -> tuple[bool, Union[dict, list, str]]:
        """Reads data from a JSON file.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, Union[dict, list, str]] -> 
                bool: True if successful, False otherwise.
                dict/list: The contents of the file if successful, error message if not.

        Calls (Without raising):
            PheonixFileNotFoundError: If file not found.
            PheonixIOError: If an input/output error occurs."""
        return super(JsonFileSystem, JsonFileSystem).ReadJsonFile(filePath)

    @staticmethod
    def DeleteJsonFile(filePath: str) -> tuple[bool, str]:
        """Deletes a JSON file.

        Args:
            filePath (str): The path to the JSON file to be deleted.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file does not exist."""
        return super(JsonFileSystem, JsonFileSystem).DeleteJsonFile(filePath)

    @staticmethod
    def FileExists(filePath: str) -> bool:
        """Checks if the JSON file exists.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            bool: True if the file exists, False otherwise."""
        return super(JsonFileSystem, JsonFileSystem).FileExists(filePath)

    @staticmethod
    def GetJsonKeys(filePath: str) -> tuple[bool, Union[list, str]]:
        """Returns all keys in a JSON file (if it is a JSON object).

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, Union[list, str]] -> 
                bool: True if successful, False otherwise.
                list: List of keys if file is a dictionary, error message if not.

        Calls (Without raising):
            PheonixFileNotFoundError: If file does not exist."""
        return super(JsonFileSystem, JsonFileSystem).GetJsonKeys(filePath)

    @staticmethod
    def GetJsonFileSize(filePath: str) -> tuple[bool, Union[int, str]]:
        """Returns the size of the JSON file.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, Union[int, str]] -> 
                bool: True if successful, False otherwise.
                int: File size in bytes if successful, error message if not.

        Calls (Without raising):
            PheonixFileNotFoundError: If file does not exist."""
        return super(JsonFileSystem, JsonFileSystem).GetJsonFileSize(filePath)

    @staticmethod
    def UpdateJsonKey(filePath: str, key: str, new_value: Union[str, int, dict, list]) -> tuple[bool, str]:
        """Updates the value of a specific key in a JSON file.

        Args:
            filePath (str): The path to the JSON file.
            key (str): The key whose value needs to be updated.
            new_value (Union[str, int, dict, list]): The new value to update.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If file does not exist."""
        return super(JsonFileSystem, JsonFileSystem).UpdateJsonKey(filePath, key, new_value)

    @staticmethod
    def ValidateJson(filePath: str) -> tuple[bool, str]:
        """Validates if a file is a valid JSON.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, str] -> 
                bool: True if valid, False otherwise.
                str: The error message if invalid.

        Calls (Without raising):
            PheonixFileNotFoundError: If file not found."""
        return super(JsonFileSystem, JsonFileSystem).ValidateJson(filePath)

    @staticmethod
    def GetJsonFileExtension(filePath: str) -> tuple[bool, str]:
        """Gets the file extension of a JSON file.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The file extension ('.json') if successful, error message if not."""
        return super(JsonFileSystem, JsonFileSystem).GetJsonFileExtension(filePath)

    @staticmethod
    def MergeJsonFiles(srcPath: str, destPath: str) -> tuple[bool, str]:
        """Merges two JSON files.

        Args:
            srcPath (str): The source JSON file path.
            destPath (str): The destination JSON file path.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful.

        Calls (Without raising):
            PheonixFileNotFoundError: If any file is not found."""
        return super(JsonFileSystem, JsonFileSystem).MergeJsonFiles(srcPath, destPath)

    @staticmethod
    def CreateJsonBackup(filePath: str) -> tuple[bool, str]:
        """Creates a backup of a JSON file.

        Args:
            filePath (str): The path to the JSON file to be backed up.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful."""
        return super(JsonFileSystem, JsonFileSystem).CreateJsonBackup(filePath)

    @staticmethod
    def RenameJsonFile(filePath: str, new_name: str) -> tuple[bool, str]:
        """Renames an existing JSON file.

        Args:
            filePath (str): The path to the existing JSON file.
            new_name (str): The new name for the file.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful."""
        return super(JsonFileSystem, JsonFileSystem).RenameJsonFile(filePath, new_name)

    @staticmethod
    def GetJsonFileInfo(filePath: str) -> tuple[bool, str]:
        """Gets the basic information (size, keys) about a JSON file.

        Args:
            filePath (str): The path to the JSON file.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful."""
        return super(JsonFileSystem, JsonFileSystem).GetJsonFileInfo(filePath)

    @staticmethod
    def ReadJsonKey(filePath: str, key: str) -> tuple[bool, Union[str, dict, list]]:
        """Reads the value of a specific key from a JSON file.

        Args:
            filePath (str): The path to the JSON file.
            key (str): The key whose value needs to be read.

        Returns:
            tuple[bool, Union[str, dict, list]] -> 
                bool: True if successful, False otherwise.
                str/dict/list: The value of the key, or error message if not found."""
        return super(JsonFileSystem, JsonFileSystem).ReadJsonKey(filePath, key)

    @staticmethod
    def SetJsonKey(filePath: str, key: str, value: Union[str, int, dict, list]) -> tuple[bool, str]:
        """Sets the value of a specific key in a JSON file.

        Args:
            filePath (str): The path to the JSON file.
            key (str): The key whose value needs to be set.
            value (Union[str, int, dict, list]): The value to set for the key.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful."""
        return super(JsonFileSystem, JsonFileSystem).SetJsonKey(filePath, key, value)

    @staticmethod
    def CreateEmptyJson(filePath: str) -> tuple[bool, str]:
        """Creates an empty JSON file.

        Args:
            filePath (str): The path where the empty JSON file will be created.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful."""
        return super(JsonFileSystem, JsonFileSystem).CreateEmptyJson(filePath)

    @staticmethod
    def CopyJsonFile(srcPath: str, destPath: str) -> tuple[bool, str]:
        """Copies a JSON file to a new location.

        Args:
            srcPath (str): The source JSON file path.
            destPath (str): The destination file path.

        Returns:
            tuple[bool, str] -> 
                bool: True if successful, False otherwise.
                str: The error message if unsuccessful."""
        return super(JsonFileSystem, JsonFileSystem).CopyJsonFile(srcPath, destPath)
    
    # 1. ASM File Handling
    @staticmethod
    def CreateAsmFile(filePath: str, asmContent: str) -> tuple[bool, str]:
        """Creates a new ASM file with the provided ASM content."""
        return super(FileSystem, FileSystem).CreateAsmFile(filePath, asmContent)

    @staticmethod
    def ReadAsmFile(filePath: str) -> tuple[bool, str]:
        """Reads content from an ASM file."""
        return super(FileSystem, FileSystem).ReadAsmFile(filePath)

    @staticmethod
    def CompileAsmToObject(filePath: str) -> tuple[bool, str]:
        """Compiles an ASM file to an object (.o) file."""
        return super(FileSystem, FileSystem).CompileAsmToObject(filePath)

    @staticmethod
    def DisassembleAsmFile(filePath: str) -> tuple[bool, str]:
        """Disassembles an ASM file to inspect the assembly instructions."""
        return super(FileSystem, FileSystem).DisassembleAsmFile(filePath)

    @staticmethod
    def CreateAsmFromString(asmCode: str, outputFile: str) -> tuple[bool, str]:
        """Creates an ASM file from a string of ASM code."""
        return super(FileSystem, FileSystem).CreateAsmFromString(asmCode, outputFile)

    # 2. Object File (.O) Handling
    @staticmethod
    def CreateObjectFile(filePath: str) -> tuple[bool, str]:
        """Creates an object (.o) file from a C/C++ file."""
        return super(FileSystem, FileSystem).CreateObjectFile(filePath)

    @staticmethod
    def LinkObjectFiles(objFiles: list[str], outputFile: str) -> tuple[bool, str]:
        """Links multiple object (.o) files into an executable."""
        return super(FileSystem, FileSystem).LinkObjectFiles(objFiles, outputFile)

    @staticmethod
    def DisassembleObjectFile(objFilePath: str) -> tuple[bool, str]:
        """Disassembles an object file to inspect the assembly instructions."""
        return super(FileSystem, FileSystem).DisassembleObjectFile(objFilePath)

    # 3. C and C++ File Handling
    @staticmethod
    def CreateCFile(filePath: str, cContent: str) -> tuple[bool, str]:
        """Creates a new C file with the provided C code."""
        return super(FileSystem, FileSystem).CreateCFile(filePath, cContent)

    @staticmethod
    def CompileCToObject(filePath: str) -> tuple[bool, str]:
        """Compiles a C file to an object (.o) file."""
        return super(FileSystem, FileSystem).CompileCToObject(filePath)

    @staticmethod
    def CreateCppFile(filePath: str, cppContent: str) -> tuple[bool, str]:
        """Creates a new C++ file with the provided C++ code."""
        return super(FileSystem, FileSystem).CreateCppFile(filePath, cppContent)

    @staticmethod
    def CompileCppToObject(filePath: str) -> tuple[bool, str]:
        """Compiles a C++ file to an object (.o) file."""
        return super(FileSystem, FileSystem).CompileCppToObject(filePath)

    @staticmethod
    def CompileCPlusPlusFile(filePath: str) -> tuple[bool, str]:
        """Compiles a C++ file into an executable."""
        return super(FileSystem, FileSystem).CompileCPlusPlusFile(filePath)

    # 4. Binary File Handling
    @staticmethod
    def ReadBinaryFile(filePath: str) -> tuple[bool, bytes]:
        """Reads data from a binary file."""
        return super(FileSystem, FileSystem).ReadBinaryFile(filePath)

    @staticmethod
    def WriteBinaryFile(filePath: str, data: bytes) -> tuple[bool, str]:
        """Writes data to a binary file."""
        return super(FileSystem, FileSystem).WriteBinaryFile(filePath, data)

    @staticmethod
    def AppendBinaryFile(filePath: str, data: bytes) -> tuple[bool, str]:
        """Appends data to a binary file."""
        return super(FileSystem, FileSystem).AppendBinaryFile(filePath, data)

    @staticmethod
    def GetBinaryFileSize(filePath: str) -> tuple[bool, int]:
        """Returns the size of a binary file."""
        return super(FileSystem, FileSystem).GetBinaryFileSize(filePath)

    # 5. Encryption and Decryption (using cryptography)
    @staticmethod
    def EncryptFile(filePath: str, key: bytes) -> tuple[bool, str]:
        """Encrypts the content of a file using AES encryption."""
        return super(FileSystem, FileSystem).EncryptFile(filePath, key)

    @staticmethod
    def DecryptFile(filePath: str, key: bytes) -> tuple[bool, str]:
        """Decrypts an encrypted file."""
        return super(FileSystem, FileSystem).DecryptFile(filePath, key)
        
    @staticmethod
    def Base64EncodeFile(filePath: str) -> tuple[bool, str]:
        """Encodes a file in Base64 format."""
        return super(FileSystem, FileSystem).Base64EncodeFile(filePath)

    @staticmethod
    def Base64DecodeFile(filePath: str) -> tuple[bool, str]:
        """Decodes a Base64 encoded file."""
        return super(FileSystem, FileSystem).Base64DecodeFile(filePath)


import os
import sys

from typing import *

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", ".."))

from phardwareitk import LIB

class CompilerOptions:
    """NOTE: Encryption is only supported in PHITKfile format."""
    def __init__(self, format_:str="COMPILER_pheonix$phardwareitk$v0001", output_format_:str="PHITKfile", encrypt:bool=True) -> None:
        """NOTE: Encryption is only supported in PHITKfile format."""
        self.format_:str = format_
        self.formatCompiler:str = format_.split("$")[0]
        self.formatModule:str = format_.split("$")[1]
        self.formatVersion:str = format_.split("$")[2]
        self.formatsCompilersAvail:list = ["COMPILER_pheonix"]
        self.formatsModulesAvail:list = ["phardwareitk"]
        self.formatsVersionsAvail:list = ["v0001", "v0002"]
        self.output_format_:str = output_format_
        self.encrypt:bool = encrypt

        if output_format_ == "PHITKfile":
            pass
        else:
            self.encrypt = False
            encrypt = False

class Compiler:
    """Main Module Handler Compiler for phardwareitk"""
    def __init__(self, path:str=LIB.Paths.defrunSettingsNOEXT, compilerOptions:CompilerOptions=CompilerOptions()) -> None:
        self.path:str = path + "." + compilerOptions.output_format_
        self.compilerOptions:CompilerOptions = compilerOptions
        self.content:str = ""
        self.sections:dict = {"_MAIN_": {}}
        self.clearedSections:dict = {"_MAIN_": {}}

        if not compilerOptions.formatCompiler in compilerOptions.formatsCompilersAvail:
            raise ValueError("In compilerOptions, Compiler Format is currently not supported or is not available. [Compiler.mainPY_ModuleController_phardwareitk.1]")
        
        if not compilerOptions.formatModule in compilerOptions.formatsModulesAvail:
            raise ValueError("In compilerOptions, Module Format is currently not supported or is not available. [Compiler.mainPY_ModuleController_phardwareitk.2]")
        
        if not compilerOptions.formatVersion in compilerOptions.formatsVersionsAvail:
            raise ValueError("In compilerOptions, Version Format is currently not supported or is not available. [Compiler.mainPY_ModuleController_phardwareitk.3]")

    def encrypt(self, data:str) -> bytes:
        import base64
        data = data.encode("utf-8")
        return base64.b64encode(data)
    
    def decrypt(self, data:bytes) -> str:
        import base64
        return base64.b64decode(data).decode("utf-8")

    def get(self, name:str, section:Union[None, str]=None) -> str:
        if section == None:
            return self.sections["_MAIN_"][name]
        else:
            return self.sections[section][name]

    def set(self, name:str, value:Union[int, str, dict], section:Union[None, str]=None) -> None:
        value = str(value)

        if section == None:
            self.sections["_MAIN_"][name] = value
        else:
            self.sections[section][name] = value

        return None 

    def makeSection(self, section:str) -> None:
        self.sections[section] = {}
        return None
    
    def decompile_v0001(self) -> dict:
        for i, v in enumerate(self.content.splitlines()):
            if v.startswith("$^S") and v.endswith("$^S"):
                self.sections[v.replace("$^S", "")] = {}

                for k, j in enumerate(self.content.splitlines()):
                    if k <= i:
                        continue

                    if " -> " in j:
                        name, value = j.split(" -> ")
                        self.sections[v.replace("$^S", "")][name] = value
            else:
                continue

        return self.sections

    def decompile_v0002(self) -> dict:
        for i, v in enumerate(self.content.splitlines()):
            if v.startswith("#%#") and v.endswith("#%#"):
                self.sections[v.replace("#%#", "")] = {}

                indexed_line = self.content.splitlines()[i + 1]
                Sindexed_line = indexed_line.split("@$@^^&^^@$@ ")

                for k, j in enumerate(Sindexed_line):
                    if " *=>>* " in j:
                        name, value = j.split(" *=>>* ")
                        self.sections[v.replace("#%#", "")][name] = value
            else:
                continue

        return self.sections

    def load(self) -> None:
        self.getFromFile()

        if self.compilerOptions.formatVersion == "v0001":
            self.decompile_v0001()
        elif self.compilerOptions.formatVersion == "v0002":
            self.decompile_v0002()
        
    def getFromFile(self) -> str:
        data:str = ""

        if self.compilerOptions.encrypt:
            with open(self.path, 'rb') as f:
                f.seek(0)
                data = self.decrypt(f.read())
                f.close()
        else:
            with open(self.path, 'r') as f:
                f.seek(0)
                data = f.read()
                f.close()

        self.content = data
        return self.content
    
    def Clear(self) -> None:
        self.sections = self.clearedSections
        self.content = ""
        return None
    
    def finalize_v0001(self) -> None:
        for i, v in enumerate(self.sections.keys()):
            if i != 0:
                self.content += "\n"
            self.content += "$^S" + v + "$^S\n"
            for k, j in enumerate(self.sections[v].keys()):
                self.content += j + " -> " + self.sections[v][j] + "\n"

    def finalize_v0002(self) -> None:
        for i, v in enumerate(self.sections.keys()):
            if i != 0:
                self.content += "\n"
            self.content += "#%#" + v + "#%#\n"
            for k, j in enumerate(self.sections[v].keys()):
                self.content += j + " *=>>* " + self.sections[v][j] + "@$@^^&^^@$@ "

    def finalize(self) -> None:
        if self.compilerOptions.formatVersion == "v0001":
            self.finalize_v0001()
        elif self.compilerOptions.formatVersion == "v0002":
            self.finalize_v0002()

    def saveToFile(self, clearAfterSave:bool=True) -> None:
        self.finalize()

        if self.compilerOptions.encrypt:
            self.content = self.encrypt(self.content)

            with open(self.path, "wb") as f:
                f.write(self.content)
                f.close()
        else:
            with open(self.path, "w") as f:
                f.write(self.content)
                f.close()

        if clearAfterSave:
            self.Clear()

        return None
    
class PHardwareITK:
    @staticmethod
    def check_file(filename:str=LIB.Paths.defrunSettingsNOEXT) -> bool:
        pass


class ELF:
    def __init__(self, labels_addr:dict, current_addr:int) -> None:
        self.labels_addr = labels_addr
        self.current_addr = current_addr
        self.text_section = bytearray()
        self.data_section = bytearray()
        self.bss_section = bytearray()
        self.function_addrs = {}
        self.no_cancelled_sections = 0
        self.string_table = bytearray(b'\0')

    def create_header(self, section_headers, debug=False):
        # ELF header
        header = bytearray()
        header.extend(b'\x7FELF')  # ELF magic number
        header.append(1)  # 32-bit
        header.append(1)  # Little-endian
        header.append(1)  # ELF version
        header.append(0)  # OS/ABI
        header.append(0)  # ABI version
        header.extend(b'\x00' * 7)  # Padding
        header.append(2)  # Type: Relocatable
        header.append(3)  # Machine: x86
        header.append(1)  # Version
        header.extend((self.current_addr).to_bytes(4, 'little'))  # Entry point
        header.extend((0).to_bytes(4, 'little'))  # Program header offset
        header.extend((0).to_bytes(4, 'little'))  # Section header offset
        header.extend((0).to_bytes(4, 'little'))  # Flags
        header.append(0x34)  # Size of this header
        header.append(0)  # Size of program header
        header.append(0)  # Number of program headers
        header.append(0x28)  # Size of section headers
        header.append(len(section_headers) // 40)  # Update for actual number of section headers
        header.append(len(section_headers) // 40 - 1)  # Section header string table index

        if debug:
            print("\nHEADERS:")
            print("\nELF:", b'\x7FELF')
            print("BITS: 1 #32-bit")
            print("ENDIAN: 1")
            print("ELF VERSION: 1")
            print("OS/ABI: 0")
            print("ABI VERISON: 0")
            print("PADDING:", b'\x00' * 7)
            print("TYPE: 2 #RELOCATABLE")
            print("MACHINE: 3 #x86")
            print("VERSION: 1")
            print("ENTRY POINT:", (self.current_addr).to_bytes(4, 'little'))
            print("PROGRAM HEADER OFFSET: ", (0).to_bytes(4, 'little'))
            print("SECTION HEADER OFFSET: ", (0).to_bytes(4, 'little'))
            print("FLAGS:", (0).to_bytes(4, 'little'))
            print("SIZE OF HEADER:", 0x34)
            print("SIZE OF PROGRAM HEADER: 0")
            print("NUMBER OF PROGRAM HEADERS: 0")
            print("SIZE OF SECTION HEADER:", 0x28)
            print("NUMBER OF SECTION HEADERS:", len(section_headers) // 40)
            print("STRING TABLE INDEX:", len(section_headers) // 40 - 1, "\n\n")

        return header

    def add_label(self, label):
        self.labels_addr[label] = self.current_addr
        self.function_addrs[label] = self.current_addr

    def advance_addr(self, size:int):
        self.current_addr += size

    def add_to_text_section(self, code):
        self.text_section.extend(code)

    def add_to_data_section(self, data):
        self.data_section.extend(data)

    def add_to_string_table(self, string:str):
        if string not in self.string_table.decode("utf-8"):
            offset = len(self.string_table)
            self.string_table.extend(string.encode('utf-8') + b'\0')
            return offset
        
        return self.string_table.decode('utf-8').index(string) + 1

    def finalize_elf(self):
        section_headers = bytearray()

        # .strtab section header
        strtab_offset = len(section_headers) + len(self.text_section) + len(self.data_section) + len(self.bss_section)
        section_headers.extend((self.add_to_string_table(".strtab")).to_bytes(4, 'little'))  # Offset in string table
        section_headers.extend((0x03).to_bytes(4, 'little'))  # Type: SHT_STRTAB
        section_headers.extend((0x00).to_bytes(4, 'little'))  # Flags
        section_headers.extend((0).to_bytes(4, 'little'))  # Address
        section_headers.extend((strtab_offset).to_bytes(4, 'little'))  # Offset in file
        section_headers.extend((len(self.string_table)).to_bytes(4, 'little'))  # Size
        section_headers.extend((0).to_bytes(4, 'little'))  # Link
        section_headers.extend((0).to_bytes(4, 'little'))  # Info
        section_headers.extend((0x01).to_bytes(4, 'little'))  # Alignment
        section_headers.extend((0).to_bytes(4, 'little'))  # Entry size

        # .text section header
        section_headers.extend((1).to_bytes(4, 'little'))  # Offset for .text in string table
        section_headers.extend((0x01).to_bytes(4, 'little'))  # Type: SHT_PROGBITS
        section_headers.extend((0x04).to_bytes(4, 'little'))  # Flags: executable
        section_headers.extend((self.current_addr).to_bytes(4, 'little'))  # Address
        section_headers.extend((0).to_bytes(4, 'little'))  # Offset in file (to be set later)
        section_headers.extend((len(self.text_section)).to_bytes(4, 'little'))  # Size
        section_headers.extend((0).to_bytes(4, 'little'))  # Link
        section_headers.extend((0).to_bytes(4, 'little'))  # Info
        section_headers.extend((0x04).to_bytes(4, 'little'))  # Alignment
        section_headers.extend((0).to_bytes(4, 'little'))  # Entry size

        # .data section header
        section_headers.extend((6).to_bytes(4, 'little'))  # Offset for .data in string table
        section_headers.extend((0x01).to_bytes(4, 'little'))  # Type: SHT_PROGBITS
        section_headers.extend((0x02).to_bytes(4, 'little'))  # Flags: writable
        section_headers.extend((self.current_addr + len(self.text_section)).to_bytes(4, 'little'))  # Address
        section_headers.extend((0).to_bytes(4, 'little'))  # Offset
        section_headers.extend((len(self.data_section)).to_bytes(4, 'little'))  # Size
        section_headers.extend((0).to_bytes(4, 'little'))  # Link
        section_headers.extend((0).to_bytes(4, 'little'))  # Info
        section_headers.extend((0x04).to_bytes(4, 'little'))  # Alignment
        section_headers.extend((0).to_bytes(4, 'little'))  # Entry size

        # .bss section header
        section_headers.extend((11).to_bytes(4, 'little'))  # Offset for .bss in string table
        section_headers.extend((0x01).to_bytes(4, 'little'))  # Type: SHT_NOBITS
        section_headers.extend((0x02).to_bytes(4, 'little'))  # Flags: writable
        section_headers.extend((self.current_addr + len(self.text_section) + len(self.data_section)).to_bytes(4, 'little'))  # Address
        section_headers.extend((0).to_bytes(4, 'little'))  # Offset
        section_headers.extend((0).to_bytes(4, 'little'))  # Size (0 for .bss)
        section_headers.extend((0).to_bytes(4, 'little'))  # Link
        section_headers.extend((0).to_bytes(4, 'little'))  # Info
        section_headers.extend((0x04).to_bytes(4, 'little'))  # Alignment
        section_headers.extend((0).to_bytes(4, 'little'))  # Entry size

        return section_headers

    def convert_to_o(self, code:str, data_vars:dict):
        opcodes = {
            "mov": {
                "eax": b"\xB8",
                "ebx": b"\xBB",
                "ecx": b"\xB9",
                "edx": b"\xBA",
                "esi": b"\xBE",
                "edi": b"\xBF",
                "ebp": b"\xBD",
                "esp": b"\xBC"
            },
            "add": {
                "eax": b"\x04",  # add eax, imm8
                "ebx": b"\x03",  # add ebx, imm32 (not direct, needs more work)
                "ecx": b"\x01",  # add ecx, imm32
                "edx": b"\x02",  # add edx, imm32
            },
            "sub": {
                "eax": b"\x2D",  # sub eax, imm32
                "ebx": b"\x3D",  # sub ebx, imm32
                "ecx": b"\x3C",  # sub ecx, imm32
                "edx": b"\x3A",  # sub edx, imm32
            },
            "int": b"\xCD",
            "call": b"\xE8"
        }
        
        code_lines = code.splitlines()

        function_addrs = {}
        
        for i, v in enumerate(code_lines):
            line = v.strip()

            if line.endswith(":"):
                self.add_label(line[:-1])
                self.add_to_text_section(f"{line[:-1]}:\n".encode('utf-8'))  # Add the function label
                continue

            code_words = line.split(" ")
            if code_words[0] in opcodes:
                if code_words[0] == "mov":
                    dest = code_words[1].replace(",", "")
                    src = code_words[2].replace(",", "")
                    
                    try:
                        im = int(src)
                        self.add_to_text_section(opcodes["mov"][dest] + im.to_bytes(4, byteorder='little'))
                    except ValueError:
                        if src in data_vars:
                            value = str(data_vars[src])
                            if not value.isdigit() and (not src.startswith('-') and not src[1:].isdigit()):
                                self.add_to_text_section(opcodes["mov"][dest] + value.encode('utf-8'))
                            else:
                                value = int(value)
                                self.add_to_text_section(opcodes["mov"][dest] + value.to_bytes(4, byteorder='little'))
                        else:
                            raise ValueError(f"Error while Compiling into '.o'. Value is not a immediate or a variable!\nLine -> {line}")
                        
                    self.advance_addr(len(opcodes["mov"][dest]) + 4)

                elif code_words[0] == "call":
                    label = code_words[1]
                    if label in function_addrs:
                        offset:int = function_addrs[label] - (self.current_addr + 5)  # 5 for E8 + rel32
                        self.add_to_text_section(opcodes["call"] + offset.to_bytes(4, byteorder='little'))
                        self.advance_addr(len(opcodes["call"]) + 4)

                elif code_words[0] == "int" and (code_words[1] == "80h" or code_words[1] == "0x80"):
                    self.add_to_text_section(opcodes["int"])
                    self.advance_addr(len(opcodes["int"]))

        return self.text_section, self.data_section



from . import *
from typing import *
import sys
import os

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

def printH(*values:object, seperator:Union[str, None]=" ", endl:Union[str, None]="\n", File:Union[str, None]=None, Flush:bool=False, backgroundColorEnabled:bool=False, FontEnabled:bool=False, Font:TextFont=TextFont()) -> None:
    """A Hyper version of Python's print function.

    Args:
        seperator (Union[str, None], optional): Seperator between values in [*values]. Defaults to " ".
        endl (Union[str, None], optional): The ending string to include at the end of the values. None is don't want. Defaults to "\n".
        File (Union[str, None], optional): Wether to save the print contents in a file before printing in console. Defaults to None.
        Flush (bool, optional): Wether to use Flush during printing. Defaults to False.
        backgroundColorEnabled (bool, optional): Wether to enable background-color. Defaults to False.
        FontEnabled (bool, optional): Wether to enable custom font. Defaults to False.
        Font (TextFont, optional): The font. Defaults to TextFont().

    NOTE: The font-size is not universally supported in all Terminals. Terminals like -> Xterm, ITerm2 (MacOS), etc do support it, but please check.
    """
    seperator = str(seperator)

    if not endl:
        endl = ""

    endl = str(endl)
    
    if not File:
        if not FontEnabled:
            i = 0
            for value in values:
                if i == 0:
                    print(str(value), end="", flush=Flush)
                else:
                    if seperator:
                        print(seperator+str(value), end="", flush=Flush)
                    else:
                        print(str(value), end="", flush=Flush)

            if endl:
                print(endl, end="", flush=Flush)
        else:
            font_code = Font.to_font_code()
            color_code = Font.color.to_rgb_code()
            backgroundColor_code = Font.background_color.to_background_code()
            resetCode = Font.to_reset_code()

            FontSize_Code = Font.to_font_size_code()

            final_str:str = ""

            if backgroundColorEnabled:
                final_str += backgroundColor_code
            if color_code:
                final_str += color_code
            if font_code:
                final_str += font_code
            if FontSize_Code:
                final_str += FontSize_Code

            i = 0
            for value in values:
                if i == 0:
                    final_str += str(value)
                else:
                    if seperator:
                        final_str += seperator + str(value)
                    else:
                        final_str += str(value)
                i += 1
            
            final_str += resetCode
            if endl:
                final_str += endl
            print(final_str, end="", flush=Flush)
    else:
        value = ""
        i = 0
        for val in values:
            if i == 0:
                value += str(val)
            else:
                value += seperator + str(val)

        if endl:
            value += endl

        with open(File, 'w') as f:
            f.write(value)
            f.close()
        
        if not FontEnabled:
            final_ = ""
            i = 0
            for value in values:
                if i == 0:
                    final_ = str(value)
                else:
                    if seperator:
                        final_ += seperator+str(value)
                    else:
                        final += str(value)

            if endl:
                final_ += endl

            print(final_, end="", flush=Flush, file=File)
        else:
            font_code = Font.to_font_code()
            color_code = Font.color.to_rgb_code()
            backgroundColor_code = Font.background_color.to_background_code()
            resetCode = Font.to_reset_code()
            FontSize_Code = Font.to_font_size_code()

            final_str:str = ""

            if backgroundColorEnabled:
                final_str += backgroundColor_code
            if color_code:
                final_str += color_code
            if font_code:
                final_str += font_code
            if FontSize_Code:
                final_str += FontSize_Code

            i = 0
            for value in values:
                if i == 0:
                    final_str += str(value)
                else:
                    if seperator:
                        final_str += seperator + str(value)
                    else:
                        final_str += str(value)
                i += 1
            
            final_str += resetCode
            if endl:
                final_str += endl
            print(final_str, end="", flush=Flush, file=File)
    
    sys.stdout.flush()

def exitH(ExitCode:Optional[int], *ExitMsg:Optional[object], seperator:Optional[str]=" ", endl:Optional[str]="\n", File:Optional[str]=None, Flush:bool=False, backgroundColorEnabled:bool=False, FontEnabled:bool=False, Font:TextFont=TextFont()) -> None:
    """A hyper version of Python's exit function.

    Args:
        ExitCode (Optional[int]): The ExitCode.
        seperator (Union[str, None], optional): Seperator between values in [*ExitMsg]. Defaults to " ".
        endl (Union[str, None], optional): The ending string to include at the end of the ExitMsg. None is don't want. Defaults to "\n".
        File (Union[str, None], optional): Wether to save the print contents in a file before printing in console. Defaults to None.
        Flush (bool, optional): Wether to use Flush during printing. Defaults to False.
        backgroundColorEnabled (bool, optional): Wether to enable background-color. Defaults to False.
        FontEnabled (bool, optional): Wether to enable custom font. Defaults to False.
        Font (TextFont, optional): The font. Defaults to TextFont().
    """
    if ExitMsg:
        printH("Status: Exit. Exit Code -> ["+str(ExitCode)+"]; Message: ", *ExitMsg, seperator=seperator, endl=endl, File=File, Flush=Flush, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)

    exit(ExitCode)

def evalH(source:Union[str, Any], globals:Optional[dict[str, Any]]=None, locals:Optional[Mapping[str, object]]=None, Log:bool=False) -> Any:
    """A Hyper version of Python's eval.

    Enhancements:
        Security: Restrics specific functions that can cause harm to OS/Files/etc. ONLY FOR STRING LITERALS.
        Log: Logging functionality.

    Args:
        source (Union[str, Any]): The source string/ReadableBuffer/CodeType object.
        globals (Optional[dict[str, Any]], optional): The globals in eval. Defaults to None.
        locals (Optional[Mapping[str, object]], optional): The locals in eval. Defaults to None.
        Log (bool, optional): Log evaluate. Defaults to False.

    Returns:
        Any: Compiled expr.
    """
    if isinstance(source, str): # Only for string literals.
        if "os." in source or "sys." in source:
            return "Cannot compile expressions with OS/Sys modules as they can/cannot harm the computer and can be used to get system info."

    if Log:
        print("Globals:", globals, "\nLocals:", locals, "\nSource:", source, "\nRunning, Eval...")

    out = eval(source, globals, locals)

    if Log:
        print("Output:", out)
    
    return out

def progressH(update:bool=False, basePos:tuple[Union[str, int]]=(0, "CPos+1"), cvalue:int=0, maxValue:int=100, useP:str="#", useNP:str="-", Pcolor:Color=Color("green"), msgColor:Color=Color("white"), NPcolor:Color=Color("red"), onMaxMsg:Optional[str]=None, defMsg:Optional[str]=None, First:bool=True, HideCursor:bool=True) -> Union[str, None]:
    """Progress Indicator Hyper.

    Args:
        update (bool, optional): Wether to update a single value, give the value to update in cvalue. Defaults to False.
        basePos (tuple[Union[str, int]], optional): The base position for the progressbar. (x, y). Use CPos to define Current Pos and you can add +/- after it to define the lines/chars. Defaults to (0, "CPos+1").
        cvalue (int, optional): The current value. Defaults to 0.
        maxValue (int, optional): The max value. Defaults to 100.
        useP (str, optional): The character to use for the parts which is already progressed. Defaults to "#".
        useNP (str, optional): The character to use for the parts which are not progressed. Defaults to "-".
        Ncolor (Color, optional): The color of the Progressed Section of the Bar. Defaults to Color("green").
        NPcolor (Color, optional): The color of the Unprogressed Section of the Bar. Defaults to Color("red").
        MsgColor (Color, optional): The color of the message. Defaults to Color("white").
        onMaxMsg (Optional[str], optional): The message to include if fully progressed. NOTE: Include ' ' before the message. Defaults to None.
        defMsg (Optional[str], optional): The default message to include after the progress Bar. Defaults to None.
        First (bool, optional): If first time called, then set to True, otherwise set to False. Defaults to True.
        HideCursor (bool, optional): If True, the cursor is hidden. Defaults to True.

    Returns:
        Union[str, None]: Str if an error occurred, otherwise None. STR 'MAX' is returned if current Value == max Value.
    """
    if cvalue > maxValue:
        return "CValue can't be greater than maxValue."
    
    x = basePos[0]
    y = basePos[1]
        
    progressBar:str = ""

    if cvalue == 0:
        progressBar = useNP * maxValue
    else:
        progressBar = useP * cvalue
        progressBar += useNP * (maxValue - cvalue)
    
    if defMsg and defMsg != "":
        progressBar += defMsg
    
    #Init cursor Pos

    from phardwareitk.CLI import cliToolKit

    if First:
        if isinstance(x, int) and isinstance(y, int):
            cliToolKit.Cursor.MoveCursor(x, y)
        elif isinstance(x, int) and not isinstance(y, int):
            cliToolKit.Cursor.MoveCursorX(x)
            if "cpos" in y.lower():
                if "+" in y.lower():
                    cliToolKit.Cursor.MoveCursorUp(y.split("+")[1])
                elif "-" in y.lower():
                    cliToolKit.Cursor.MoveCursorDown(y.split("-")[1])
        elif not isinstance(x, int) and isinstance(y, int):
            cliToolKit.Cursor.MoveCursor(0, y)
            if "cpos" in x.lower():
                if "+" in x.lower():
                    cliToolKit.Cursor.MoveCursorRight(x.split("+")[1])
                elif "-" in x.lower():
                    cliToolKit.Cursor.MoveCursorLeft(x.split("-")[1])
        else:
            if "cpos" in y.lower():
                if "+" in y.lower():
                    cliToolKit.Cursor.MoveCursorUp(y.split("+")[1])
                elif "-" in y.lower():
                    cliToolKit.Cursor.MoveCursorDown(y.split("-")[1])

            if "cpos" in x.lower():
                if "+" in x.lower():
                    cliToolKit.Cursor.MoveCursorRight(x.split("+")[1])
                elif "-" in x.lower():
                    cliToolKit.Cursor.MoveCursorLeft(x.split("-")[1])

    if HideCursor:
        cliToolKit.Cursor.HideCursor()

    if update:
        cliToolKit.Cursor.MoveCursorRight(cvalue - 1)
        printH(useNP, FontEnabled=True, endl="", Font=TextFont(font_color=NPcolor))
        return None
    
    for char in progressBar:
        color = None
        if char == useNP:
            color = NPcolor
        else:
            color = Pcolor
        printH(char, FontEnabled=True, endl="", Font=TextFont(font_color=color))

    if cvalue == maxValue:
        if onMaxMsg and onMaxMsg != "":
            printH(onMaxMsg, FontEnabled=True, endl="", Font=TextFont(font_color=msgColor))

    if cvalue != maxValue or (not onMaxMsg or onMaxMsg == ""):
        cliToolKit.Cursor.MoveCursorLeft(maxValue)
    else:
        cliToolKit.Cursor.MoveCursorRight(maxValue - len(onMaxMsg))
        print("\n")
    
    if cvalue == maxValue:
        cliToolKit.Cursor.ShowCusor()

    if cvalue == maxValue:
        return "MAX"
    
    return None

def cacheH(path:str, callable:Optional[object]=None, expiration:Optional[int]=None) -> Union[Optional[str], Any]:
    """Cache Hyper function. 
    
    Generates Cache files, Incase file already exists, returns the function output.


    Useful For:

        1. Large Functions that cannot be run easily.

        2. Cache functions.

        3. Large Data.

        4. Easy Loading and Unloading.

    Args:
        path (str): The Path to the Cache file.
        callable (Optional[object], optional): The function to cache. Defaults to None.
        expiration (Optional[int], optional): If not None, The expiration time, in hours. Defaults to None.

    Returns:
        None | str | Any: string, for success, if string includes '$Error$' at the start there is a Error. In case file already exists, the result of the callable function will be returned.
    """

    import time

    if expiration:
        expiration = int(expiration) * 3600 # Covert to seconds:

    cache_data:Any = None

    if os.path.exists(path):
        try:
            with open(path, 'rb') as f:
                cache_data = f.read()

            pbin = PBin(header=None)
            stringD, CBFlag, CB_data = pbin.Deserialize(cache_data) # Deserialize

            #Extract timestamp
            timestamp = CB_data.split('##SEPERATOR##')[1]
            result = CB_data.split('##SEPERATOR##')[0]

            nameSpace = {}
            exec(result, nameSpace)
            funcName = result.split("(")[0].split()[-1]
            result = nameSpace[funcName]

            if not expiration:
                return result

            # If expriation, check if it is expired or not.
            if time.time() - timestamp < expiration:
                return result
            
        except Exception as e:
            return f"$Error$ Error loading cache data. -> {e}"
        
    import inspect

    result = inspect.getsource(callable)

    cTimeStamp = int(time.time())
    cache_data_ = result + f"##SEPERATOR##{cTimeStamp}"

    pbin = PBin(header=PHeader())
    binResult = pbin.Serialize(cache_data_, CB_flag=True)

    with open(path, 'wb') as f:
        f.write(binResult)

    return result


from typing import *

################################################################

# PHardwareITK Errors

class PHardwareITKException(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PHardwareITK_RuntimeError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PHardwareITK_NotSupportedError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PHardwareITK_ValueError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PHardwareITK_ParserError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PHardwareITK_SettingsError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PHardwareITK_FileError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PHardwareITK_FileNotFoundError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PHardwareITK_IOError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

# Pheonix Errors

class PheonixException(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixFileNotFoundError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixInternalError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixValuerError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixArgumentError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixSettingsError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixObjectError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixOsError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixOsVersionError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixVersionNotSupported(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixUnknownVersion(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixUnknownSetting(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

class PheonixIOError(Exception):
    def __init__(self, message:str="", *args) -> None:
        super().__init__(message)

# Custom Errors

class CustomError(Exception):
    """Custom Error
    
    Args:
        name (str): The name of the Exception.
        message (str): The message displayed.
        *args (Any): Other Arguments.
        
    Calls:
        Exception"""
    def __init__(self, name:str, message:str="", *args) -> None:
        self.name = name
        self.message = message

        super().__init__(message)

    def __str__(self):
        return f"{self.name}: {self.message}"

class RunCustomError(): # Same as CustomError
    """Run Custom Error
    
    Args:
        name (str): The name of the Exception.
        message (str): The message displayed.
        *args (Any): Other Arguments.
        
    Calls:
        CustomError -> Exception
        
    NOTE: Same as Custom Error."""
    def __init__(self, name:str, message:str="", *args) -> None:
        raise CustomError(name, message, *args)


"""GUI Library using SDL.

NOTE: Still under development.
"""

from . import *

if __name__ == '__main__':
    exit()

import os
import sys

import ctypes

from typing import *

from sdl2 import *

from . import *

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.ErrorSystem.ErrorSystem import *
from phardwareitk.Extensions import *

class WidgetPack():
    """A type for a single PACKED Widget.
    """
    def __init__(self, Widget:PWidget, MainFunc:object, MainFuncParams:list) -> None:
        """INITALIZE func.

        Args:
            Widget (PWidget): The Widget class.
            MainFunc (function): The main function of the widget.
            MainFuncParams (list): The parameters of the main function.
        """
        self.Widget = Widget
        self.MainFunc = MainFunc
        self.MainFuncParams = MainFuncParams

def initialize() -> None:
    """Initializes the SDL and PHardware GUI.

    Raises:
        PheonixException: Incase SDL does not initialize properly.
    """
    if SDL_Init(SDL_INIT_VIDEO) != 0:
        raise PheonixException("Could not initialize SDL. - ["+str(SDL_GetError())+"]")
    
def SetBackgroundColor(window: Any, color:Color=Color("white")) -> None:
    """Sets the background color of the specified window.

    Args:
        window (Any): The window
        color (Color, optional): The color. [phardwareitk.Extensions.Color]. Defaults to Color("white").
    """
    windowSurface = SDL_GetWindowSurface(window)
    if not windowSurface:
        return None
    
    r,g,b = color.color

    sdl_color = SDL_Color(r=r, g=g, b=b, a=color.alpha)
    SDL_FillRect(windowSurface, None, SDL_MapRGBA(windowSurface.contents.format, sdl_color.r, sdl_color.g, sdl_color.b, sdl_color.a))
    SDL_UpdateWindowSurface(window)

def SetSurfaceBackgroundColor(surface:Any, color:Color=Color("white")) -> None:
    """Sets the background color of the specified surface.

    Args:
        surface (Any): The surface
        color (Color, optional): The color. [phardwareitk.Extensions.Color]. Defaults to Color("white").
    """
    if not surface:
        return None
    
    r,g,b = color.color
    
    sdl_color = SDL_Color(r=r, g=g, b=b, a=color.alpha)
    SDL_FillRect(surface, None, SDL_MapRGBA(surface.contents.format, sdl_color.r, sdl_color.g, sdl_color.b, sdl_color.a))
    
def AddIcon(window:Any, icon:PIcon) -> None:
    """Adds an icon to the window.

    Args:
        window (Any): Window.
        icon (PIcon): Icon. [phardwareitk.Extensions.PIcon]
    """
    icon.SetIconSDL2(window)
    
def WindowSurface(window:Any) -> Any:
    """Returns the SDL2 window surface.

    Args:
        window (Any): The window returned by 'CreateWindow' Function.

    Returns:
        Any: Window surface.
    """
    return SDL_GetWindowSurface(window)

def LoadBMP(image:str="") -> Any:
    """Loads the specified BMP image.

    Args:
        image (str, optional): The image path in string. Defaults to "".

    Returns:
        Any: The returned value of SDL2.'LoadBMP' Function.
    """
    return SDL_LoadBMP(ToBytes(image))

def UpdateWindow(window:Any) -> None:
    """Updates the window.

    Args:
        window (Any): The window.
    """
    SDL_UpdateWindowSurface(window)

def BlitSurfaceScaled(src:Any, srcrect:Any, dst:Any, dstrect:Any, scaleMode:Any) -> bool:
    """_summary_

    Args:
        src (Any): the SDL_Surface structure to be copied from.
        srcrect (Any): the SDL_Rect structure representing the rectangle to be copied, or NULL to copy the entire surface.
        dst (Any): the SDL_Surface structure that is the blit target.
        dstrect (Any): the SDL_Rect structure representing the target rectangle in the destination surface, or NULL to fill the entire destination surface.
        scaleMode (Any): the SDL_ScaleMode to be used.
    """
    return SDL_BlitScaled

def BlitSurface(src:Any, srcrect:Any, dst:Any, dstrect:Any) -> bool:
    """Same as SDL2.BlitSurface.

    Args:
        src (Any): the SDL_Surface structure to be copied from.
        srcrect (Any): the SDL_Rect structure representing the rectangle to be copied, or NULL to copy the entire surface.
        dst (Any): the SDL_Surface structure that is the blit target.
        dstrect (Any): the SDL_Rect structure representing the x and y position in the destination surface, or NULL for (0,0). The width and height are ignored, and are copied from srcrect. If you want a specific width and height, you should use BlitSurfaceScaled().
    """
    return SDL_BlitSurface(src, srcrect, dst, dstrect)


def FreeSurface(image:Any) -> None:
    """Frees the specified BMP image.

    Args:
        image (Any): The image returned from 'LoadBMP' function.
    """
    SDL_FreeSurface(image)
    
def CreateWindow(name:str="PHardware GUI", width:int=800, height:int=400, WindowPos:Union[None, Any]=SDL_WINDOWPOS_CENTERED) -> Any:
    """Creates a window.

    Args:
        name (str, optional): The Title of the window. Defaults to "PHardware GUI".
        width (int, optional): The width of the window. Defaults to 800.
        height (int, optional): The height of the window. Defaults to 400.
        WindowPos (Union[None, Any], optional): The position of the window. Only accept SDL_WINDOWPOS_[Something]. Defaults to SDL_WINDOWPOS_CENTERED

    Raises:
        PheonixException: Incase window creation failed.

    Returns:
        Any: Window Handle
    """

    if WindowPos == None:
        WindowPos = SDL_WINDOWPOS_CENTERED

    name = ToBytes(name)

    window = SDL_CreateWindow(name, WindowPos, WindowPos, width, height, SDL_WINDOW_SHOWN)
    if not window:
        raise PheonixException("Could not create window. - ["+str(SDL_GetError())+"]")
    else:
        return window
    
def DestroyWindow(window:Any, icon:Optional[PIcon]=None, renderer:Optional[SDL_Renderer]=None) -> None:
    """Destroyes the specified window.

    Args:
        window (Any): The window.
        icon (Optional[PIcon]): If used, the icon provided in 'AddIcon' function.
        renderer (Optional[SDL_Renderer]): If used, the renderer provided in MakeRenderer, RenderText function
    """
    SDL_DestroyWindow(window)
    if icon:
        icon.Free()
    
    if renderer:
        SDL_DestroyRenderer(renderer)

def Quit() -> None:
    """Quits the Window.
    """
    SDL_Quit()
    
def EventLoop(quit_OnMessage:bool=False, window:Any=None, WidgetPacks:Optional[list[WidgetPack]]=None, icon:Optional[PIcon]=None, renderer:Optional[SDL_Renderer]=None) -> Union[bool, SDL_Event]:
    """Checks if any events are in event queue and reteives the next event.

    Args:
        quit_OnMessage (bool, optional): If set to True, will run loop (Button and such widgets will not work), upon receiving Quit event (e.g. On Window Quit) the program will destroy the window. Else False. Defaults to False
        window (Any, optional): If quit_OnMessage set to true, the provide the window
        WidgetPacks (Optional[list[WidgetPack]]): If any, the program will provide events to the widget.
        ....: Parameter to DestroyWindow.
        
    Returns:
        Union[bool, SDL_Event]: True for Quit, otherwise the event.
    """
    event = SDL_Event()

    run = True
    while run:
        while SDL_PollEvent(ctypes.byref(event)) != 0:
            if WidgetPacks != None and WidgetPacks != []:
                for i, v in enumerate(WidgetPacks):
                    if not v.MainFunc:
                        continue

                    if v.MainFuncParams == None or v.MainFuncParams == []:
                        v.MainFunc(event)
                    else:
                        params = tuple(v.MainFuncParams)
                        v.MainFunc(event, *params)
                
            if event.type == SDL_QUIT:
                if quit_OnMessage:
                    if not icon and not renderer:
                        DestroyWindow(window)
                    elif icon and not renderer:
                        DestroyWindow(window, icon)
                    elif not icon and renderer:
                        DestroyWindow(window, renderer=renderer)
                    elif icon and renderer:
                        DestroyWindow(window, icon, renderer)

                    Quit()
                else:
                    return True
                run = False
                break

def MakeRenderer(window:Any, index:int=-1, flags:Any=0) -> Any:
    """Returns a renderer.

    Args:
        window (Any): The Window.
        index (int, optional): The index of the rendering device to activate. -1 for default. Defaults to -1.
        flags (Any, optional): The flags for Renderer.

    Returns:
        Any: The renderer.
    """
    return SDL_CreateRenderer(window, index, flags)

def RenderText(window:Any, widget_rect:SDL_Rect, text:str, font:TextFont=TextFont(font_color=Color("black")), fontFile:Optional[str]=None) -> Union[SDL_Renderer, bytes]:
    """Renders text.

    Args:
        window (Any): The window to render the text in.
        widget_rect (SDL_Rect): The rectangle inside which the text should be rendered. (RECT will be invisible)
        text (str): The text to be rendered.
        font (TextFont): The text font to be rendered.
        fontFile (Optional[str], optional): The Font file, if not present -> None. Defaults to None._
    """
    from sdl2 import sdlttf
    if sdlttf.TTF_Init() > 0:
        return None
    
    render_ = MakeRenderer(window)
    r,g,b = font.color.color
    SDL_SetRenderDrawColor(render_, r,g,b,font.color.alpha)
    SDL_RenderClear(render_)

    font_ = None

    if Windows and not fontFile:
        font_ = sdlttf.TTF_OpenFont(ToBytes(f"C:\\Windows\\Fonts\\{font.font}.ttf"), 64)
    elif Linux and not fontFile:
        font_ = sdlttf.TTF_OpenFont(ToBytes(f"usr/share//fonts//{font.font}.ttf"), 64)
    elif Unix and not fontFile:
        font_ = sdlttf.TTF_OpenFont(ToBytes(f"usr/share/fonts//{font.font}.ttf"), 64)
    elif Unknown_os and not fontFile:
        return b"Unknown OS"
    
    if not font_:
        return ToBytes(FromBytes(SDL_GetError()) + " !PHardware -> font == None")

    if fontFile:
        font_ = sdlttf.TTF_OpenFont(ToBytes(fontFile), 64)

    color = SDL_Color(r, g, b, font.color.alpha)
    tsurface = sdlttf.TTF_RenderText_Solid(font_, ToBytes(text), color)  

    if not tsurface: return ToBytes(FromBytes(SDL_GetError()) + " !PHardware -> Tsurface == None")

    ttexture = SDL_CreateTextureFromSurface(render_, tsurface)

    if not ttexture: return ToBytes(FromBytes(SDL_GetError()) + " !PHardware -> Ttexture == None")

    SDL_FreeSurface(tsurface)

    SDL_RenderCopy(render_, ttexture, None, widget_rect)
    SDL_RenderPresent(render_)

    SDL_DestroyTexture(ttexture)

    return render_

class Button(PWidget):
    """A Button widget

    Args:
        PWidget (PWidget): phardwareitk.Extensions.PWidget
    """

    def __init__(self, x:int, y:int, width:int, height:int, label:str="ClickMe", bg_color:Color=Color("gray"), Font:TextFont=TextFont()) -> None:
        """Initialize the Button.

        Args:
            x (int): X coordinate.
            y (int): Y coordinate.
            width (int): Width.
            height (int): Height.
            label (str, optional): Label for the button. Defaults to "Button".
            bg_color (Color, optional): Background-Color. Defaults to Color("gray").
            Font (TextFont, optional): Font for the label. Defaults to TextFont().
        """

        super().__init__(x, y, width, height, 0, bg_color, True, Font, label)

        self.callback:object = None
        self.params:tuple = None

    def Draw(self, window:Any) -> SDL_Renderer:
        """Draws the button widget.

        Args:
            window (Any): The window to draw the widget.
        """
        surface = WindowSurface(window)
        pixel_format = surface.contents.format
        r,g,b = self.bgColor.color
        mapped_color = SDL_MapRGBA(pixel_format, r, g, b, self.bgColor.alpha)

        SDL_FillRect(surface, SDL_Rect(self.x, self.y, self.width, self.height), mapped_color)

        return RenderText(window, SDL_Rect(self.x, self.y, self.width, self.height), self.text, self.textFont) if self.text else None

    def onClick(self, callback:object, params:Optional[tuple]=None) -> WidgetPack:
        """Runs the following function if Mouse is clicked upon the button.

        Args:
            callback (object): The callback function.
            params (Optional[tuple], optional): If any, parameters for the callback. Defaults to None.

        Returns:
            WidgetPack: Use it inside the EventLoop.
        """
        self.callback = callback
        self.params = params

        return WidgetPack(Button, lambda event: self._onevent(event), [])
    
    def _onevent(self, event:SDL_Event) -> None:
        """Event handler. DO NOT CALL DIRECTLY. RECOMMENDED TO CALL VIA EVENT_LOOP.
        """
        if event.type == SDL_MOUSEBUTTONDOWN:
            mx = ctypes.c_int(0)
            my = ctypes.c_int(0)
            SDL_GetMouseState(ctypes.byref(mx), ctypes.byref(my))

            mx = mx.value
            my = my.value

            if self.x <= mx <= self.x + self.width and self.y <= my <= self.y + self.height:
                if self.callback:
                    if not self.params:
                        self.callback()
                    else:
                        self.callback(*self.params)